"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const buffer_hexdump_1 = tslib_1.__importDefault(require("buffer-hexdump"));
const events_1 = require("events");
const Connect_1 = require("./messages/Connect");
const Ping_1 = require("./messages/Ping");
const Publish_1 = require("./messages/Publish");
const PublishAck_1 = require("./messages/PublishAck");
const PublishRecorded_1 = require("./messages/PublishRecorded");
const Subscribe_1 = require("./messages/Subscribe");
const Unsubscribe_1 = require("./messages/Unsubscribe");
const MqttConnection_1 = tslib_1.__importDefault(require("./MqttConnection"));
const MqttPacket_1 = require("./MqttPacket");
const MqttTypes_1 = require("./MqttTypes");
const debug_1 = tslib_1.__importDefault(require("debug"));
const debugLog = debug_1.default("fblib");
class MqttApiEmitter extends events_1.EventEmitter {
}
/**
 * Handles decoding and sending all sorts of messages used by Facebook Messenger.
 * It utilizes all network primitives defined in the MqttConnection class.
 */
class MqttApi {
    constructor() {
        this._connected = false;
        this.lastMsgId = 1;
        this.emitter = new MqttApiEmitter();
        this.sendPing = () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.connection.writeMessage(Ping_1.encodePing());
            setTimeout(this.sendPing, 60 * 1000);
        });
        this.connection = new MqttConnection_1.default();
    }
    on(event, cb) {
        this.emitter.on(event, cb);
    }
    sendSubscribe(msg) {
        this.lastMsgId = this.lastMsgId + 1;
        return this.connection.writeMessage(msg);
    }
    connect() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.connection.connect();
            this._connected = true;
        });
    }
    /**
     * Sends a CONNECT mqtt message and binds listeners for recieving messages.
     * @param tokens
     * @param deviceId
     */
    sendConnectMessage(tokens, deviceId) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this.tokens = tokens;
            setTimeout(this.sendPing, 60 * 1000);
            if (!this._connected) {
                return;
            }
            this.connection.emitter.on("packet", (packet) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                switch (packet.type) {
                    case MqttPacket_1.FacebookMessageType.ConnectAck:
                        debugLog("Packet type: ConnectAck");
                        yield this.sendPublish("/foreground_state", '{"foreground":true,"keepalive_timeout":60}');
                        yield this.sendSubscribe(Subscribe_1.encodeSubscribeMessage(this.lastMsgId));
                        yield this.sendSubscribe(Unsubscribe_1.encodeUnsubscribe(this.lastMsgId));
                        debugLog("Connected.");
                        // await this.sendMessage()
                        this.emitter.emit("connected");
                        break;
                    case MqttPacket_1.FacebookMessageType.Publish:
                        debugLog("Packet type: Publish");
                        const publish = Publish_1.decodePublish(packet);
                        this.emitter.emit("publish", publish);
                        this.sendPublishConfirmation(packet.flag, publish);
                        break;
                    case MqttPacket_1.FacebookMessageType.SubscribeAck:
                        debugLog("Packet type: SubscribeAck");
                        break;
                    case MqttPacket_1.FacebookMessageType.PublishAck:
                        debugLog("Packet type: PublishAck");
                        break;
                    case MqttPacket_1.FacebookMessageType.UnsubscribeAck:
                        debugLog("Packet type: UnsubscribeAck");
                        break;
                    case MqttPacket_1.FacebookMessageType.Pong:
                        debugLog("Packet type: Pong");
                        break;
                    default:
                        debugLog("Packet type:", packet.type);
                        debugLog(buffer_hexdump_1.default(packet.content));
                }
            }));
            const connectMessage = yield Connect_1.encodeConnectMessage(tokens, deviceId);
            yield this.connection.writeMessage(connectMessage);
        });
    }
    sendPublish(topic, data) {
        const packet = Publish_1.encodePublish(this.lastMsgId, topic, data);
        this.lastMsgId += 1;
        return this.connection.writeMessage(packet);
    }
    /**
     * Sends a facebook messenger message to someone.
     */
    sendMessage(threadId, message) {
        return new Promise((resolve, reject) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const milliseconds = Math.floor(new Date().getTime() / 1000);
            const rand = Math.floor(Math.random() * (Math.pow(2, 32) - 1));
            const msgid = (rand & 0x3fffff) | (milliseconds << 22);
            const msg = {
                body: message,
                msgid,
                sender_fbid: this.tokens.uid,
                to: threadId
            };
            yield this.sendPublish("/send_message2", JSON.stringify(msg));
            const sentMessageInfoListener = (publish) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                if (publish.topic !== "/t_ms")
                    return;
                const content = JSON.parse(publish.content.toString("utf8").replace("\u0000", ""));
                if (!content.deltas || !content.deltas.length)
                    return;
                const [delta] = content.deltas;
                if (!delta.deltaSentMessage)
                    return;
                resolve(delta.deltaSentMessage);
                this.emitter.removeListener("publish", sentMessageInfoListener);
            });
            this.on("publish", sentMessageInfoListener);
        }));
    }
    sendPublishConfirmation(flags, publish) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (publish.msgId === 0) {
                return;
            }
            const qos1 = (flags & MqttTypes_1.MqttMessageFlag.QoS1) == MqttTypes_1.MqttMessageFlag.QoS1;
            const qos2 = (flags & MqttTypes_1.MqttMessageFlag.QoS2) == MqttTypes_1.MqttMessageFlag.QoS2;
            if (qos1) {
                this.connection.writeMessage(PublishAck_1.encodePublishAck(publish.msgId));
            }
            if (qos2) {
                this.connection.writeMessage(PublishRecorded_1.encodePublishRecorded(publish.msgId));
            }
        });
    }
}
exports.default = MqttApi;
//# sourceMappingURL=MqttApi.js.map